"""
Miscellaneous utilities
"""

import ctypes
import warnings
from functools import wraps
from inspect import Parameter, signature
from typing import Any, Callable, TypeVar

import numpy as np

_CTYPES_TYPE_TABLE = {
    "uint32": ctypes.c_uint32,
    "float32": ctypes.c_float,
    "float64": ctypes.c_double,
}


_NUMPY_TYPE_TABLE = {"uint32": np.uint32, "float32": np.float32, "float64": np.float64}


_PyBytes_FromStringAndSize = ctypes.pythonapi.PyBytes_FromStringAndSize
_PyBytes_FromStringAndSize.argtypes = (ctypes.c_char_p, ctypes.c_ssize_t)
_PyBytes_FromStringAndSize.restype = ctypes.py_object


def typestr_to_ctypes_type(type_info):
    """Obtain ctypes type corresponding to a given Type str"""
    return _CTYPES_TYPE_TABLE[type_info]


def typestr_to_numpy_type(type_info):
    """Obtain ctypes type corresponding to a given Type str"""
    return _NUMPY_TYPE_TABLE[type_info]


def c_str(string):
    """Convert a Python string to C string"""
    return ctypes.c_char_p(string.encode("utf-8"))


def bytes_from_string_and_size(ptr, size):
    """Copy `size` bytes from `ptr` to create a new python `bytes` object"""
    # Theoretically `ctypes.string_at` does this, but the `size` argument
    # there only takes an `int`, while python bytes object can support up to a
    # `ssize_t` in size.
    return _PyBytes_FromStringAndSize(ptr, size)


def py_str(string):
    """Convert C string back to Python string"""
    return string.decode("utf-8")


def c_array(ctype, values):
    """
    Convert a Python byte array to C array

    WARNING
    -------
    DO NOT USE THIS FUNCTION if performance is critical. Instead, use np.array(*)
    with dtype option to explicitly convert type and then use
    ndarray.ctypes.data_as(*) to expose underlying buffer as C pointer.
    """
    return (ctype * len(values))(*values)


_T = TypeVar("_T")


# Notice for `require_keyword_args`
# Authors: Olivier Grisel
#          Gael Varoquaux
#          Andreas Mueller
#          Lars Buitinck
#          Alexandre Gramfort
#          Nicolas Tresegnie
#          Sylvain Marie
# License: BSD 3 clause
def _require_keyword_args(
    error: bool,
) -> Callable[[Callable[..., _T]], Callable[..., _T]]:
    """Decorator for methods that issues warnings for positional arguments

    Using the keyword-only argument syntax in pep 3102, arguments after the
    * will issue a warning or error when passed as a positional argument.

    Modified from sklearn utils.validation.

    Parameters
    ----------
    error :
        Whether to throw an error or raise a warning.
    """

    def throw_if(func: Callable[..., _T]) -> Callable[..., _T]:
        """Throw an error/warning if there are positional arguments after the asterisk.

        Parameters
        ----------
        func :
            function to check arguments on.

        """
        sig = signature(func)
        kwonly_args = []
        all_args = []

        for name, param in sig.parameters.items():
            if param.kind == Parameter.POSITIONAL_OR_KEYWORD:
                all_args.append(name)
            elif param.kind == Parameter.KEYWORD_ONLY:
                kwonly_args.append(name)

        @wraps(func)
        def inner_f(*args: Any, **kwargs: Any) -> _T:
            extra_args = len(args) - len(all_args)
            if not all_args and extra_args > 0:  # keyword argument only
                raise TypeError("Keyword argument is required.")

            if extra_args > 0:
                # ignore first 'self' argument for instance methods
                args_msg = [
                    f"{name}"
                    for name, _ in zip(kwonly_args[:extra_args], args[-extra_args:])
                ]
                # pylint: disable=consider-using-f-string
                msg = "Pass `{}` as keyword args.".format(", ".join(args_msg))
                if error:
                    raise TypeError(msg)
                warnings.warn(msg, FutureWarning)
            for k, arg in zip(sig.parameters, args):
                kwargs[k] = arg
            return func(**kwargs)

        return inner_f

    return throw_if


deprecate_positional_args = _require_keyword_args(False)



"""Converter to export Treelite models as scikit-learn models (EXPERIMENTAL)"""

from enum import IntEnum
from typing import Any

import numpy as np

from ..core import TreeliteError
from ..model import Model


def _ensure_scalar_int(x: Any) -> int:
    if isinstance(x, np.ndarray):
        assert x.shape == (1,)
        return int(x[0])
    try:
        return int(x)
    except ValueError as e:
        raise ValueError(f"Cannot interpret x as a scalar integer, {x.type=}") from e


def _ensure_numpy(x: Any) -> np.ndarray:
    if isinstance(x, np.ndarray):
        return x
    raise ValueError(f"x is not a valid NumPy array. {x.type=}")


_node_dtype = np.dtype(
    {
        "names": [
            "left_child",
            "right_child",
            "feature",
            "threshold",
            "impurity",
            "n_node_samples",
            "weighted_n_node_samples",
            "missing_go_to_left",
        ],
        "formats": ["<i8", "<i8", "<i8", "<f8", "<f8", "<i8", "<f8", "u1"],
        "offsets": [0, 8, 16, 24, 32, 40, 48, 56],
        "itemsize": 64,
    }
)


class _TaskType(IntEnum):
    # pylint: disable=invalid-name
    kBinaryClf = 0
    kRegressor = 1
    kMultiClf = 2
    kLearningToRank = 3
    kIsolationForest = 4


def _export_tree(
    model, *, tree_id, n_features, n_classes, n_targets, tree_depths, subestimator_class
):
    # pylint: disable=too-many-locals
    try:
        from sklearn import __version__ as sklearn_version
        from sklearn.tree import DecisionTreeClassifier
        from sklearn.tree._tree import Tree as SKLearnTree
    except ImportError as e:
        raise TreeliteError("This function requires scikit-learn package") from e

    tree_accessor = model.get_tree_accessor(tree_id)
    has_categorical_split = tree_accessor.get_field("has_categorical_split").tolist()[0]
    if has_categorical_split:
        raise NotImplementedError(
            "Trees with categorical splits cannot yet be exported as scikit-learn"
        )

    tree = SKLearnTree(n_features, n_classes, n_targets)

    n_nodes = tree_accessor.get_field("num_nodes").tolist()[0]
    nodes = np.empty(n_nodes, dtype=_node_dtype)

    nodes["left_child"] = tree_accessor.get_field("cleft")
    nodes["right_child"] = tree_accessor.get_field("cright")
    nodes["feature"] = tree_accessor.get_field("split_index")
    nodes["threshold"] = tree_accessor.get_field("threshold")
    nodes["impurity"] = np.nan
    nodes["n_node_samples"] = -1
    nodes["weighted_n_node_samples"] = np.nan
    nodes["missing_go_to_left"] = tree_accessor.get_field("default_left")

    if n_targets == 1 and n_classes[0] == 1:
        leaf_value = (
            tree_accessor.get_field("leaf_value").astype("float64").reshape((-1, 1, 1))
        )
    else:
        # Need to map leaf values to correct layout
        leaf_value = np.zeros((n_nodes, n_targets, n_classes[0]), dtype="float64")
        leaf_value_raw = tree_accessor.get_field("leaf_vector").astype("float64")
        leaf_vec_begin = tree_accessor.get_field("leaf_vector_begin")
        leaf_vec_end = tree_accessor.get_field("leaf_vector_end")
        for node_id in range(n_nodes):
            if leaf_vec_begin[node_id] != leaf_vec_end[node_id]:
                # This node is a leaf node and outputs a vector
                leaf_value[node_id, :, :] = leaf_value_raw[
                    leaf_vec_begin[node_id] : leaf_vec_end[node_id]
                ].reshape((n_targets, n_classes[0]))

    state = {
        "max_depth": tree_depths[tree_id],
        "node_count": n_nodes,
        "nodes": nodes,
        "values": leaf_value,
    }
    tree.__setstate__(state)

    subestimator = subestimator_class()
    subestimator_state = {
        "tree_": tree,
        "n_outputs_": n_targets,
        "_sklearn_version": sklearn_version,
    }
    if subestimator_class is DecisionTreeClassifier:
        if n_targets == 1:
            subestimator_state["n_classes_"] = n_classes[0]
        else:
            subestimator_state["n_classes_"] = n_classes.tolist()
    subestimator.__setstate__(subestimator_state)
    return subestimator


def export_model(model: Model) -> Any:
    """
    Export a model as a scikit-learn RandomForest.

    Note
    ----
    Currently only random forests can be exported as scikit-learn model objects.
    Support for gradient boosted trees and other kinds of tree models will be
    added in the future.

    Parameters
    ----------
    model : :py:class:`Model`
        Treelite mobel to export

    Returns
    -------
    sklearn_model : object of type \
                    :py:class:`~sklearn.ensemble.RandomForestRegressor` / \
                    :py:class:`~sklearn.ensemble.RandomForestClassifier` / \
                    :py:class:`~sklearn.ensemble.GradientBoostingRegressor` / \
                    :py:class:`~sklearn.ensemble.GradientBoostingClassifier`
        Scikit-learn model
    """
    # pylint: disable=too-many-locals
    try:
        from sklearn import __version__ as sklearn_version
        from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
        from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
    except ImportError as e:
        raise TreeliteError("This function requires scikit-learn package") from e

    header_accessor = model.get_header_accessor()
    average_tree_output = (
        _ensure_scalar_int(header_accessor.get_field("average_tree_output")) == 1
    )
    task_type = _ensure_scalar_int(header_accessor.get_field("task_type"))
    n_features = _ensure_scalar_int(header_accessor.get_field("num_feature"))
    n_trees = _ensure_scalar_int(header_accessor.get_field("num_tree"))
    n_targets = _ensure_scalar_int(header_accessor.get_field("num_target"))
    n_classes = _ensure_numpy(header_accessor.get_field("num_class"))
    leaf_vector_shape = _ensure_numpy(header_accessor.get_field("leaf_vector_shape"))
    target_id = _ensure_numpy(header_accessor.get_field("target_id"))
    class_id = _ensure_numpy(header_accessor.get_field("class_id"))
    tree_depths = model.get_tree_depth()

    # Heuristics to ensure that the model can be represented as scikit-learn random forest
    # 1. average_tree_output must be True
    # 2. n_classes[i] must be identical for all targets
    # 3. Each leaf must yield an output of shape (n_targets, n_classes)
    # 4. target_id[i] must be either 0 or -1
    # 5. class_id[i] must be either 0 or -1
    def raise_not_rf_error(reason):
        raise NotImplementedError(
            "This Treelite model cannot be represented as scikit-learn random forest. "
            f"Condition unmet: {reason}"
            "Other kinds of tree models in scikit-learn are not yet supported."
        )

    if not average_tree_output:
        raise_not_rf_error(
            "Outputs of tree outputs must be averaged to produce the final output"
        )
    if not np.all(n_classes == n_classes[0]):
        raise_not_rf_error("n_classes must be identical for all trees")
    if not np.array_equal(leaf_vector_shape, [n_targets, n_classes.max()]):
        raise_not_rf_error(
            "Each tree must produce output of dimensions (n_targets, n_classes)"
        )
    if not np.all((target_id == 0) | (target_id == -1)):
        raise_not_rf_error("target_id field must be either 0 or -1")
    if not np.all((class_id == 0) | (class_id == -1)):
        raise_not_rf_error("class_id field must be either 0 or -1")

    if task_type in [_TaskType.kBinaryClf, _TaskType.kMultiClf]:
        estimator_class = RandomForestClassifier
        subestimator_class = DecisionTreeClassifier
    else:
        estimator_class = RandomForestRegressor
        subestimator_class = DecisionTreeRegressor

    estimators = []

    for tree_id in range(n_trees):
        estimators.append(
            _export_tree(
                model,
                tree_id=tree_id,
                n_features=n_features,
                n_classes=n_classes,
                n_targets=n_targets,
                tree_depths=tree_depths,
                subestimator_class=subestimator_class,
            )
        )

    clf = estimator_class()
    state = {
        "estimators_": estimators,
        "n_outputs_": n_targets,
        "n_features_in_": n_features,
        "_sklearn_version": sklearn_version,
    }
    if estimator_class is RandomForestClassifier:
        if n_targets == 1:
            state.update(
                {
                    "n_classes_": n_classes[0],
                    "classes_": np.arange(n_classes[0]),
                }
            )
        else:
            state.update(
                {
                    "n_classes_": n_classes.tolist(),
                    "classes_": [np.arange(n_classes[i]) for i in range(n_targets)],
                }
            )
    clf.__setstate__(state)

    return clf



"""Functions to load and build model objects"""

from __future__ import annotations

import ctypes
import json
import pathlib
from typing import Any, Union

from . import compat
from .core import _LIB, _check_call
from .model import Model
from .util import c_str, py_str


def _normalize_path(filename: Union[str, pathlib.Path]) -> pathlib.Path:
    """
    Fully expand a path and convert it to an absolute path
    """
    path = pathlib.Path(filename)
    return path.expanduser().resolve()


def load_xgboost_model_legacy_binary(filename: Union[str, pathlib.Path]) -> Model:
    """
    Load a tree ensemble model from XGBoost model, stored using
    the legacy binary format. Note: new XGBoost models should
    be stored in the JSON format, to take advantage of the
    latest functionalities of XGBoost.

    Parameters
    ----------
    filename :
        Path to model file

    Returns
    -------
    model : :py:class:`Model`
        Loaded model

    Example
    -------

    .. code-block:: python

       xgb_model = treelite.frontend.load_xgboost_model_legacy_binary(
           "xgboost_model.model")
    """
    path = _normalize_path(filename)
    return Model(handle=compat.load_xgboost_model_legacy_binary(str(path)))


def load_xgboost_model(
    filename: Union[str, pathlib.Path],
    *,
    format_choice: str = "use_suffix",
    allow_unknown_field: bool = False,
) -> Model:
    """
    Load a tree ensemble model from XGBoost model, stored using JSON or UBJSON format.

    Parameters
    ----------
    filename :
        Path to model file
    format_choice :
        Method to select the model format

        * ``use_suffix`` (default): Use the suffix of the file name (also known as file
          extension) to detect the format. Files whose names end with ``.json`` will be
          parsed as JSON; all other files will be parsed as UBJSON.
        * ``inspect``: Inspect the first few bytes of the file to heuristically determine
          whether the file is JSON or UBJSON.
        * ``ubjson``: Parse the file as UBJSON.
        * ``json``: Parse the file as JSON.
    allow_unknown_field:
        Whether to allow extra fields with unrecognized keys

    Returns
    -------
    model : :py:class:`Model`
        Loaded model

    Example
    -------

    .. code-block:: python

       xgb_model = treelite.frontend.load_xgboost_model("xgboost_model.json")
    """

    parser_config = {"allow_unknown_field": allow_unknown_field}
    parser_config_str = json.dumps(parser_config)

    path = _normalize_path(filename)

    def parse_as_json() -> Model:
        return Model(
            handle=compat.load_xgboost_model(
                str(path), allow_unknown_field=allow_unknown_field
            )
        )

    def parse_as_ubjson() -> Model:
        handle = ctypes.c_void_p()
        _check_call(
            _LIB.TreeliteLoadXGBoostModelUBJSON(
                c_str(str(path)), c_str(parser_config_str), ctypes.byref(handle)
            )
        )
        return Model(handle=handle)

    if format_choice == "use_suffix":
        if path.name.endswith(".json"):
            return parse_as_json()
        # File name not ending with .json will be parsed as UBJSON.
        return parse_as_ubjson()

    if format_choice == "inspect":
        detected_format = _detect_xgboost_format(path)
        if detected_format == "json":
            return parse_as_json()
        if detected_format == "ubjson":
            return parse_as_ubjson()
        raise ValueError(
            "Could not detect whether the given XGBoost model is JSON or UBJSON. "
            "Please explicitly set format_choice='json' or format_choice='ubjson'"
        )
    if format_choice == "ubjson":
        return parse_as_ubjson()

    if format_choice == "json":
        return parse_as_json()

    raise ValueError(f"Unknown format_choice argument: {format_choice}")


def load_lightgbm_model(filename: Union[str, pathlib.Path]) -> Model:
    """
    Load a tree ensemble model from a LightGBM model file.

    Parameters
    ----------
    filename :
        Path to model file

    Returns
    -------
    model : :py:class:`Model`
        Loaded model

    Example
    -------

    .. code-block:: python

       lgb_model = treelite.frontend.load_lightgbm_model("lightgbm_model.txt")
    """
    path = _normalize_path(filename)
    return Model(handle=compat.load_lightgbm_model(str(path)))


def from_xgboost(booster: Any) -> Model:
    """
    Load a tree ensemble model from an XGBoost Booster object

    Parameters
    ----------
    booster : Object of type :py:class:`xgboost.Booster`
        Python handle to XGBoost model

    Returns
    -------
    model : :py:class:`Model`
        Loaded model
    """
    return Model(handle=compat.from_xgboost(booster))


def from_xgboost_json(
    model_json_str: Union[bytes, bytearray, str],
    *,
    allow_unknown_field: bool = False,
) -> Model:
    """
    Load a tree ensemble model from a string containing XGBoost JSON

    Parameters
    ----------
    model_json_str :
        A string specifying an XGBoost model in the XGBoost JSON format
    allow_unknown_field:
        Whether to allow extra fields with unrecognized keys

    Returns
    -------
    model: :py:class:`Model`
        Loaded model
    """
    return Model(
        handle=compat.from_xgboost_json(
            model_json_str, allow_unknown_field=allow_unknown_field
        )
    )


def from_xgboost_ubjson(
    model_ubjson_str: Union[bytes, bytearray],
    *,
    allow_unknown_field: bool = False,
) -> Model:
    """
    Load a XGBoost model from a byte sequence containing UBJSON

    Parameters
    ----------
    model_ubjson_str :
        A byte sequence specifying an XGBoost model in the UBJSON format
    allow_unknown_field:
        Whether to allow extra fields with unrecognized keys

    Returns
    -------
    model: :py:class:`Model`
        Loaded model
    """
    return Model(
        handle=compat.from_xgboost_ubjson(
            model_ubjson_str, allow_unknown_field=allow_unknown_field
        )
    )


def from_lightgbm(booster: Any) -> Model:
    """
    Load a tree ensemble model from a LightGBM Booster object

    Parameters
    ----------
    booster : object of type :py:class:`lightgbm.Booster`
        Python handle to LightGBM model

    Returns
    -------
    model : :py:class:`Model`
        Loaded model
    """
    return Model(handle=compat.from_lightgbm(booster))


def _detect_xgboost_format(path: pathlib.Path) -> str:
    """
    Inspect the first few bytes of an XGBoost model and heuristically determine whether
    it's using the JSON or UBJSON format.

    Parameters
    ----------
    path :
        Path to model file (assumed to be fully resolved with ``path.expanduser().resolve()``)

    Returns
    -------
    "json", "ubjson", or "unknown"
    """
    detected_format = ctypes.c_char_p()
    _check_call(
        _LIB.TreeliteDetectXGBoostFormat(
            c_str(str(path)),
            ctypes.byref(detected_format),
        )
    )
    return py_str(detected_format.value)


__all__ = [
    "load_xgboost_model_legacy_binary",
    "load_xgboost_model",
    "load_lightgbm_model",
    "from_xgboost",
    "from_xgboost_json",
    "from_xgboost_ubjson",
    "from_lightgbm",
]
